import pytest
from unittest.mock import MagicMock

from typing import Callable
from importlib.metadata import EntryPoint
from pathlib import Path

from mocks import (
    MockMotor, 
    MockMotorInfo, 
    MockDetector, 
    MockDetectorInfo,
    MockController,
    MockControllerInfo,
)

@pytest.fixture
def config_path() -> Path:
    return Path(__file__).parent / "mocks" / "configs"

@pytest.fixture
def mock_detector_entry_points() -> Callable[[str], list[EntryPoint]]:

    def mocked_detector_entry_points(group: str) -> list[EntryPoint]:
        plugins = []

        if group == "redsun.plugins.models.config":
            info_ep = EntryPoint(
                name="detector",
                value="mock_detector:MockDetectorInfo",
                group="redsun.plugins.models.config",
            )
            object.__setattr__(info_ep, "load", MagicMock(return_value=MockDetectorInfo))
            plugins.append(info_ep)

        if group == "redsun.plugins.models":
            model_ep = EntryPoint(
                name="detector",
                value="mock_detector:MockDetector",
                group="redsun.plugins.models",
            )
            object.__setattr__(model_ep, "load", MagicMock(return_value=MockDetector))
            plugins.append(model_ep)

        return plugins
    
    return mocked_detector_entry_points

@pytest.fixture
def mock_motor_entry_points() -> Callable[[str], list[EntryPoint]]:

    def mocked_motor_entry_points(group: str) -> list[EntryPoint]:
        plugins = []

        if group == "redsun.plugins.models.config":
            info_ep = EntryPoint(
                name="motor",
                value="mock_motor:MockMotorInfo",
                group="redsun.plugins.models.config",
            )
            object.__setattr__(info_ep, "load", MagicMock(return_value=MockMotorInfo))
            plugins.append(info_ep)

        if group == "redsun.plugins.models":
            model_ep = EntryPoint(
                name="motor", 
                value="mock_motor:MockMotor", 
                group="redsun.plugins.models"
            )
            object.__setattr__(model_ep, "load", MagicMock(return_value=MockMotor))
            plugins.append(model_ep)

        return plugins
    
    return mocked_motor_entry_points

@pytest.fixture
def mock_controller_entry_points() -> Callable[[str], list[EntryPoint]]:

    def mocked_controller_entry_points(group: str) -> list[EntryPoint]:
        plugins = []

        if group == "redsun.plugins.controllers.config":
            info_ep = EntryPoint(
                name="controller",
                value="mock_controller:MockControllerInfo",
                group="redsun.plugins.controllers.config",
            )
            object.__setattr__(info_ep, "load", MagicMock(return_value=MockControllerInfo))
            plugins.append(info_ep)
        if group == "redsun.plugins.controllers":
            model_ep = EntryPoint(
                name="controller",
                value="mock_controller:MockController",
                group="redsun.plugins.controllers",
            )
            object.__setattr__(model_ep, "load", MagicMock(return_value=MockController))
            plugins.append(model_ep)
        
        return plugins
    return mocked_controller_entry_points



# type: ignore
import pytest

from unittest.mock import MagicMock, patch

import logging
from pathlib import Path
from importlib.metadata import EntryPoint
from typing import Callable
from redsun.controller.plugins import PluginManager
from sunflare.config import RedSunSessionInfo

from mocks import (
    mocked_motor_missing_entry_points, 
    mocked_motor_mismatched_entry_points, 
    mocked_motor_non_derived_info_entry_points,
    mocked_ctrl_non_derived_info_entry_points,
    mocked_ctrl_non_derived_entry_points,
    mocked_ctrl_mismatched_entry_points
)

logging.basicConfig(level=logging.DEBUG)

def test_load_motor_plugins(config_path: Path, mock_motor_entry_points: Callable[[str], list[EntryPoint]]) -> None:
    # Create a mock that returns our function
    mock_ep = MagicMock(side_effect=mock_motor_entry_points)

    # Need to patch where entry_points is imported, not where it's defined
    with patch('redsun.controller.plugins.entry_points', mock_ep):

        # Then test through the plugin manager
        config, types_groups = PluginManager.load_configuration(str(config_path / "mock_motor_config.yaml"))

        assert isinstance(config, RedSunSessionInfo)
        assert len(types_groups["models"]) == 2
        assert ["Single axis motor", "Double axis motor"] == list(types_groups["models"].keys())

        # check config for single axis motor
        assert config.models["Single axis motor"].model_name == "MockMotor"
        assert config.models["Single axis motor"].axis == ["X"]
        assert config.models["Single axis motor"].integer == 42
        assert config.models["Single axis motor"].floating == 3.14
        assert config.models["Single axis motor"].string == "Hello, World!"

        # check config for double axis motor
        assert config.models["Double axis motor"].model_name == "MockMotor"
        assert config.models["Double axis motor"].axis == ["X", "Y"]
        assert config.models["Double axis motor"].egu == "mm"
        assert config.models["Double axis motor"].integer == 314
        assert config.models["Double axis motor"].floating == 4.2
        assert config.models["Double axis motor"].string == "Goodbye, World!"

def test_load_detector_plugins(config_path: Path, mock_detector_entry_points: Callable[[str], list[EntryPoint]]) -> None:
    # Create a mock that returns our function
    mock_ep = MagicMock(side_effect=mock_detector_entry_points)

    # Need to patch where entry_points is imported, not where it's defined
    with patch('redsun.controller.plugins.entry_points', mock_ep):

        # Then test through the plugin manager
        config, types_groups = PluginManager.load_configuration(str(config_path / "mock_detector_config.yaml"))

        assert isinstance(config, RedSunSessionInfo)
        assert len(config.models) == 2
        assert len(types_groups["models"]) == 2
        assert ["iSCAT channel", "TIRF channel"] == list(types_groups["models"].keys())

def test_load_controller_plugins(config_path: Path, mock_controller_entry_points: Callable[[str], list[EntryPoint]]) -> None:
    # Create a mock that returns our function
    mock_ep = MagicMock(side_effect=mock_controller_entry_points)

    # Need to patch where entry_points is imported, not where it's defined
    with patch('redsun.controller.plugins.entry_points', mock_ep):

        # Then test through the plugin manager
        config, types_groups = PluginManager.load_configuration(str(config_path / "mock_controller_config.yaml"))

        assert isinstance(config, RedSunSessionInfo)
        assert len(config.controllers) == 1
        assert len(config.models) == 0
        assert len(types_groups["controllers"]) == 1
        assert ["MockController"] == list(types_groups["controllers"].keys())

mocked_error_entry_points = [
    mocked_motor_mismatched_entry_points,
    mocked_motor_missing_entry_points,
    mocked_motor_non_derived_info_entry_points,
    mocked_ctrl_non_derived_info_entry_points,
    mocked_ctrl_non_derived_entry_points,
    mocked_ctrl_mismatched_entry_points
]

@pytest.mark.parametrize("mock_entry_points", mocked_error_entry_points)
def test_errors_plugin_loading(config_path: Path, mock_entry_points: Callable[[str], list[EntryPoint]]) -> None:
    # Create a mock that returns our function
    mock_ep = MagicMock(side_effect=mock_entry_points)

    # Need to patch where entry_points is imported, not where it's defined
    with patch('redsun.controller.plugins.entry_points', mock_ep):
        config, types_groups = PluginManager.load_configuration(str(config_path / "mock_motor_config.yaml"))
        assert config.models == {}
        assert config.controllers == {}
        assert len(types_groups["models"]) == 0
        assert len(types_groups["controllers"]) == 0



"""Entry points used to test error conditions."""

import sys
from unittest.mock import MagicMock

from .mock_detector import *
from .mock_motor import *
from .mock_controller import *

if sys.version_info >= (3, 10):
    from importlib.metadata import EntryPoint
else:
    from importlib_metadata import EntryPoint

def mocked_motor_missing_entry_points(group: str) -> list[EntryPoint]:
    plugins = []

    if group == "redsun.plugins.models":
        model_ep = EntryPoint(
            name="motor",
            value="mock_motor:MockMotor",
            group="redsun.plugins.models",
        )
        object.__setattr__(model_ep, "load", MagicMock(return_value=MockMotor))
        plugins.append(model_ep)
    return plugins

def mocked_motor_mismatched_entry_points(group: str) -> list[EntryPoint]:
    plugins = []
    if group == "redsun.plugins.models.config":
        info_ep = EntryPoint(
            name="motor_config",
            value="mock_motor:MockMotorInfo",
            group="redsun.plugins.models.config",
        )
        plugins.append(info_ep)
    if group == "redsun.plugins.models":
        model_ep = EntryPoint(
            name="motor",
            value="mock_motor:MockMotor",
            group="redsun.plugins.models",
        )
        plugins.append(model_ep)
    return plugins

def mocked_motor_non_derived_info_entry_points(group: str) -> list[EntryPoint]:
    plugins = []
    if group == "redsun.plugins.models.config":
        info_ep = EntryPoint(
            name="motor",
            value="mock_motor:NonDerivedDetectorInfo",
            group="redsun.plugins.models.config",
        )
        object.__setattr__(info_ep, "load", MagicMock(return_value=NonDerivedMotorInfo))
        plugins.append(info_ep)

    if group == "redsun.plugins.models":
        model_ep = EntryPoint(
            name="motor",
            value="mock_motor:MockMotor",
            group="redsun.plugins.models",
        )
        object.__setattr__(model_ep, "load", MagicMock(return_value=MockMotor))
        plugins.append(model_ep)
    return plugins

def mocked_ctrl_mismatched_entry_points(group: str) -> list[EntryPoint]:
    plugins = []
    if group == "redsun.plugins.controllers.config":
        info_ep = EntryPoint(
            name="controller_config",
            value="mock_controller:MockControllerInfo",
            group="redsun.plugins.controllers.config",
        )
        plugins.append(info_ep)
    if group == "redsun.plugins.controllers":
        model_ep = EntryPoint(
            name="controller",
            value="mock_controller:MockController",
            group="redsun.plugins.controllers",
        )
        plugins.append(model_ep)
    return plugins

def mocked_ctrl_non_derived_info_entry_points(group: str) -> list[EntryPoint]:
    plugins = []
    if group == "redsun.plugins.controllers.config":
        info_ep = EntryPoint(
            name="controller",
            value="mock_controller:NonDerivedMotorInfo",
            group="redsun.plugins.controllers.config",
        )
        plugins.append(info_ep)

    if group == "redsun.plugins.controllers":
        model_ep = EntryPoint(
            name="controller",
            value="mock_controller:MockController",
            group="redsun.plugins.controllers",
        )
        plugins.append(model_ep)
    return plugins

def mocked_ctrl_non_derived_entry_points(group: str) -> list[EntryPoint]:
    plugins = []
    if group == "redsun.plugins.controllers.config":
        info_ep = EntryPoint(
            name="controller",
            value="mock_controller:MockControllerInfo",
            group="redsun.plugins.controllers.config",
        )
        plugins.append(info_ep)
    if group == "redsun.plugins.controllers":
        model_ep = EntryPoint(
            name="controller",
            value="mock_controller:NonDerivedController",
            group="redsun.plugins.controllers",
        )
        plugins.append(model_ep)
    return plugins



"""Redsun main entry point."""

from __future__ import annotations

import argparse
import sys

from sunflare.virtual import VirtualBus

from redsun.controller import PluginManager, RedsunController
from redsun.view import build_view_layer, create_app, launch_app


class RedSunArgs(argparse.Namespace):
    """type hints for command line arguments."""

    config: str


def parse_args() -> RedSunArgs:
    """Parse command line arguments.

    Returns
    -------
    RedSunArgs
        Parsed command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Redsun: event-driven data acquisition system"
    )
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        help="Absolute path to the YAML configuration file",
        required=True,
    )

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args: RedSunArgs = parser.parse_args(namespace=RedSunArgs())
    return args


def main(input_config: str) -> None:
    """Redsun application entry point.

    Parameters
    ----------
    input_config : str
        Path to the configuration file.
    """
    # create the application
    app = create_app()

    # virtual layer
    virtual_bus = VirtualBus()

    # get the startup configuration
    config, types_groups = PluginManager.load_configuration(input_config)

    # build the controller layer
    controller = RedsunController(config, virtual_bus, types_groups)

    # build the view layer
    view = build_view_layer(config, types_groups["widgets"], virtual_bus)

    # connect the controller and the view to the virtual layer
    controller.connect_to_virtual()
    view.connect_to_virtual()

    # launch the application;
    # the app starts here and
    # there is no return until it's closed
    launch_app(app, view)


def main_cli() -> None:
    """Command line entry point for the Redsun application."""
    args = parse_args()
    main(args.config)


if __name__ == "__main__":
    main_cli()



from .mock_detector import MockDetector, MockDetectorInfo
from .mock_motor import MockMotor, MockMotorInfo, NonDerivedMotor, NonDerivedMotorInfo
from .mock_controller import MockController, MockControllerInfo, NonDerivedController, NonDerivedControllerInfo
from .mock_entry_points import (
    mocked_motor_missing_entry_points, 
    mocked_motor_mismatched_entry_points, 
    mocked_motor_non_derived_info_entry_points,
    mocked_ctrl_non_derived_info_entry_points,
    mocked_ctrl_non_derived_entry_points,
    mocked_ctrl_mismatched_entry_points
)

__all__ = [
    "MockDetector",
    "MockDetectorInfo",
    "MockMotor",
    "MockMotorInfo",
    "NonDerivedMotor",
    "NonDerivedMotorInfo",
    "MockController",
    "MockControllerInfo",
    "NonDerivedController",
    "NonDerivedControllerInfo",
    "mocked_motor_missing_entry_points",
    "mocked_motor_mismatched_entry_points",
    "mocked_motor_non_derived_info_entry_points",
    "mocked_ctrl_non_derived_info_entry_points",
    "mocked_ctrl_non_derived_entry_points",
    "mocked_ctrl_mismatched_entry_points"
]



from __future__ import annotations

from typing import Any

from attrs import define, field, validators, setters, Attribute

from collections import OrderedDict

from sunflare.config import ModelInfo
from sunflare.engine import Status
from sunflare.model import ModelProtocol

from bluesky.protocols import Reading
from event_model.documents.event_descriptor import DataKey

@define(kw_only=True)
class MockDetectorInfo(ModelInfo):
    """Mock motor model information."""

    exposure: float = field(validator=validators.instance_of(float))
    egu: str = field(
        default="s", validator=validators.instance_of(str), on_setattr=setters.frozen
    )
    sensor_shape: tuple[int, int] = field(converter=tuple, on_setattr=setters.frozen)
    pixel_size: tuple[float, float, float] = field(converter=tuple)
    integer: int
    floating: float
    string: str

    @sensor_shape.validator
    def _validate_sensor_shape(
        self, _: Attribute[tuple[int, ...]], value: tuple[int, ...]
    ) -> None:
        if not all(isinstance(val, int) for val in value):
            raise ValueError("All values in the tuple must be integers.")
        if len(value) != 2:
            raise ValueError("The tuple must contain exactly two values.")
    @pixel_size.validator
    def _validate_pixel_size(
        self, _: Attribute[tuple[float, ...]], value: tuple[float, ...]
    ) -> None:
        if not all(isinstance(val, float) for val in value):
            raise ValueError("All values in the tuple must be floats.")
        if len(value) != 3:
            raise ValueError("The tuple must contain exactly three values.")

class MockDetector:
    def __init__(self, name: str, model_info: MockDetectorInfo) -> None:
        self.name = name
        self.model_info = model_info

    def stage(self) -> Status:
        raise NotImplementedError(
            "Mock detector model does not support stage operation."
        )

    def unstage(self) -> Status:
        raise NotImplementedError(
            "Mock detector model does not support unstage operation."
        )

    def read(self) -> OrderedDict[str, Reading[Any]]:
        raise NotImplementedError(
            "Mock detector model does not support read operation."
        )

    def shutdown(self) -> None:
        raise NotImplementedError(
            "Mock detector model does not support shutdown operation."
        )

    def pause(self) -> None:
        raise NotImplementedError(
            "Mock detector model does not support pause operation."
        )

    def resume(self) -> None:
        """Perform device-specific work when the RunEngine resumes after a pause."""
        raise NotImplementedError(
            "Mock detector model does not support resume operation."
        )

    def kickoff(self) -> Status:
        raise NotImplementedError(
            "Mock detector model does not support kickoff operation."
        )

    def complete(self) -> Status:
        raise NotImplementedError(
            "Mock detector model does not support complete operation."
        )

    def read_configuration(self) -> OrderedDict[str, Reading[Any]]:
        raise NotImplementedError(
            "Mock detector model does not support read_configuration operation."
        )

    def describe_configuration(self) -> OrderedDict[str, DataKey]:
        raise NotImplementedError(
            "Mock detector model does not support describe_configuration operation."
        )

    @property
    def integer(self) -> int:
        return self.model_info.integer

    @property
    def floating(self) -> float:
        return self.model_info.floating

    @property
    def string(self) -> str:
        return self.model_info.string


class OtherDetector(ModelProtocol):
    def __init__(self, name: str, model_info: MockDetectorInfo) -> None:
        super().__init__(name, model_info)

    def read(self) -> OrderedDict[str, Reading[Any]]:
        raise NotImplementedError("Mock detector model does not support read operation.")
    
    def stage(self) -> Status:
        raise NotImplementedError("Mock detector model does not support stage operation.")

    def unstage(self) -> Status:
        raise NotImplementedError("Mock detector model does not support unstage operation.")
    
    def read_configuration(self) -> OrderedDict[str, Reading[Any]]:
        raise NotImplementedError("Mock detector model does not support read_configuration operation.")

    def describe_configuration(self) -> OrderedDict[str, DataKey]:
        raise NotImplementedError("Mock detector model does not support describe_configuration operation.")
    
    def kickoff(self) -> Status:
        raise NotImplementedError("Mock detector model does not support kickoff operation.")
    
    def complete(self) -> Status:
        raise NotImplementedError("Mock detector model does not support complete operation.")
    
    def pause(self) -> None:
        raise NotImplementedError("Mock detector model does not support pause operation.")
    
    def resume(self) -> None:
        raise NotImplementedError("Mock detector model does not support resume operation.")

    def shutdown(self) -> None: ...

    @property
    def integer(self) -> int:
        return self.model_info.integer

    @property
    def floating(self) -> float:
        return self.model_info.floating

    @property
    def string(self) -> str:
        return self.model_info.string



from typing import ClassVar

from attrs import define

from sunflare.config import ControllerInfo
from sunflare.controller import ControllerProtocol
from sunflare.model import ModelProtocol
from sunflare.virtual import VirtualBus
from psygnal import SignalGroupDescriptor

@define(kw_only=True)
class MockControllerInfo(ControllerInfo):
    """Mock controller information model."""
    string: str
    integer: int
    floating: float
    boolean: bool


class MockController(ControllerProtocol):

    _ctrl_info: MockControllerInfo

    def __init__(self, 
                ctrl_info: MockControllerInfo, 
                models: dict[str, ModelProtocol],
                virtual_bus: VirtualBus):
        self._ctrl_info = ctrl_info
        self._models = models
        self._virtual_bus = virtual_bus

    def shutdown(self) -> None:
        ...

    def registration_phase(self) -> None:
        ...

    def connection_phase(self) -> None:
        ...

class NonDerivedControllerInfo:
    """Non-derived controller information model."""
    string: str
    integer: int
    floating: float
    boolean: bool

class NonDerivedController:

    def __init__(self, 
                ctrl_info: MockControllerInfo, 
                models: dict[str, ModelProtocol],
                virtual_bus: VirtualBus):
            ...
        
    def shutdown(self) -> None:
        ...

    def registration_phase(self) -> None:
        ...

    def connection_phase(self) -> None:
        ...



from __future__ import annotations

from typing import Any

from attrs import define, field, setters, validators

from sunflare.config import ModelInfo
from sunflare.model import ModelProtocol
from sunflare.engine import Status

from bluesky.protocols import Location

@define(kw_only=True)
class MockMotorInfo(ModelInfo):
    """Mock motor model information."""

    axis: list[str] = field(factory=list, on_setattr=setters.frozen)
    step_size: dict[str, float] = field(factory=dict)
    egu: str = field(validator=validators.instance_of(str), on_setattr=setters.frozen)
    integer: int
    floating: float
    string: str

    @axis.validator
    def _validate_axis(self, _: str, value: list[str]) -> None:
        if not all(isinstance(val, str) for val in value):
            raise ValueError("All values in the list must be strings.")
        if len(value) == 0:
            raise ValueError("The list must contain at least one element.")
    @step_size.validator
    def _validate_step_size(self, _: str, value: dict[str, float]) -> None:
        if not all(isinstance(val, float) for val in value.values()):
            raise ValueError("All values in the dictionary must be floats.")
        if len(value) == 0:
            raise ValueError("The dictionary must contain at least one element.")

class NonDerivedMotorInfo:
    """Mock detector model information."""

    axis: list[str]
    step_size: dict[str, float]
    egu: str
    integer: int
    floating: float
    string: str

class MockMotor(ModelProtocol):

    def __init__(self, name: str, model_info: MockMotorInfo) -> None:
        self.name = name
        self.model_info = model_info

    def set(self, value: int, *args: Any, **kwargs: Any) -> Status:  # type: ignore[override]
        raise NotImplementedError("Mock motor model does not support set operation.")

    def locate(self) -> Location[int]:  # type: ignore[override]
        raise NotImplementedError("Mock motor model does not support locate operation.")

    def shutdown(self) -> None: ...

    @property
    def model_info(self) -> MockMotorInfo:
        return self.model_info

    @property
    def integer(self) -> int:
        return self.model_info.integer

    @property
    def floating(self) -> float:
        return self.model_info.floating

    @property
    def string(self) -> str:
        return self.model_info.string
    

class NonDerivedMotor:

    def __init__(self, name: str, model_info: NonDerivedMotorInfo) -> None:
        self.name = name
        self.model_info = model_info

    def set(self, value: int, *args: Any, **kwargs: Any) -> Status:
        raise NotImplementedError("Mock motor model does not support set operation.")
    
    def locate(self) -> Location[int]:
        raise NotImplementedError("Mock motor model does not support locate operation.")
    
    def shutdown(self) -> None: ...
    
    @property
    def integer(self) -> int:
        return self.model_info.integer
    
    @property
    def floating(self) -> float:
        return self.model_info.floating
    
    @property
    def string(self) -> str:
        return self.model_info.string


class OtherMotor(ModelProtocol):
    def __init__(self, name: str, model_info: MockMotorInfo) -> None:
        self.name = name
        self.model_info = model_info

    def set(self, value: int, *args: Any, **kwargs: Any) -> Status:  # type: ignore[override]
        raise NotImplementedError("Mock motor model does not support set operation.")

    def locate(self) -> Location[int]:  # type: ignore[override]
        raise NotImplementedError("Mock motor model does not support locate operation.")
    
    def shutdown(self) -> None: ...
    
    @property
    def integer(self) -> int:
        return self.model_info.integer
    
    @property
    def floating(self) -> float:
        return self.model_info.floating
    
    @property
    def string(self) -> str:
        return self.model_info.string
